/*
	void GetMatrix(nx, ny, xl, xr, yb, yt, sk, id, ik)

	Computes matrix for Poisson eqution. 
	a) Determine coordinates and element connectivity matrix wrt. 
		an equidistant discretization.
	b) We use a general code to calculate element matrices and 
		local stiffness matrix K. 

	IN:	nx, ny	points in x/y direction
		dx, dy  subdomain-size in x/y-direction
	OUT:	sk	symmetric matrix
		id, ik	crs pointer of sk (symmetric matrix)
*/
#include <iostream>
#include <cmath>
using namespace std;
#include "header.h"
/* 
	numbering: rowise
*/
void GetCoordsInRectangle(const int nx, const int ny, 
	const double& xl, const double& xr, const double& yb, const double& yt, 
	double xc[])
{ 
  const double hx = (xr-xl)/nx,
               hy = (yt-yb)/ny;
  
  int k = 0;
  for (int j=0; j<=ny; ++j)
  {
    const double y0 = yb + j*hy;
    for (int i=0; i<=nx; ++i, k+=2)
    {
      xc[k  ] = xl + i*hx;
      xc[k+1] = y0;
    }
  }

  return;
}

/* 
	element numbering: rowise - lower/upper
	node numbering in element: lex. forward
*/
// void GetConnectivityInRectangle(const int nx, const int ny, int ia[][3])
void GetConnectivityInRectangle(const int nx, const int ny, int ia[])
{
  const int dx = nx+1;
        int k  = 0;
        int l  = 0;
  for (int j=0; j<ny; ++j, ++k)
  {
    for (int i=0; i<nx; ++i, ++k)
    {
      ia[l  ] = k;
      ia[l+1] = k+1;
      ia[l+2] = k+dx+1;
      l+=3;
      ia[l  ] = k;
      ia[l+1] = k+dx;
      ia[l+2] = k+dx+1;
      l+=3;     
    } 
  }
  return;
}

/* 
	symmetric matrix stored in CRS, end of row pointer
*/
void GetMatrixpointersInRectangle(const int nx, const int ny, int id[], int ik[])
{
  const int dx = nx+1;
        int k  = 0;					/*  node counter */
        int ij = -1;					/* entry counter */
  for (int j=0; j<ny; ++j)
  {
    for (int i=0; i<nx; ++i, ++k)
    {
      ++ij;
      ik[ij] = k;
      ++ij;
      ik[ij] = k+1;
      ++ij;
      ik[ij] = k+dx;
      ++ij;
      ik[ij] = k+dx+1;
      id[k]  = ij;
    }
     
      ++ij;					/* right edge of domain */
      ik[ij] = k;
      ++ij;
      ik[ij] = k+dx;
      id[k]  = ij; 
      ++k; 
  }
  						/*   top edge of domain */
  for (int i=0; i<nx; ++i, ++k)
  {
    ++ij;
    ik[ij] = k;
    ++ij;
    ik[ij] = k+1;
    id[k]  = ij;
  } 
  					/*   top right corner of domain */
    ++ij;
    ik[ij] = k;
    id[k]  = ij;

  return;
}

/* 
			general routine for lin. triangular elements
*/
// void CalcElem(const int ial[3], const double xc[][2], double ske[3][3], double fe[3])
void CalcElem(const int ial[3], const double xc[], double ske[3][3], double fe[3])
{
//   extern double FunctF(double x,double y);
  const int  i1  = 2*ial[0],   i2= 2*ial[1],   i3 = 2*ial[2];
  const double x13 = xc[i3+0] - xc[i1+0],  y13 = xc[i3+1] - xc[i1+1],
               x21 = xc[i1+0] - xc[i2+0],  y21 = xc[i1+1] - xc[i2+1],
               x32 = xc[i2+0] - xc[i3+0],  y32 = xc[i2+1] - xc[i3+1];
  const double jac = fabs(x21*y13-x13*y21); 
  
  ske[0][0] = 0.5/jac* (y32*y32+x32*x32);
  ske[0][1] = 0.5/jac* (y13*y32+x13*x32);
  ske[0][2] = 0.5/jac* (y21*y32+x21*x32);
  ske[1][0] = ske[0][1];
  ske[1][1] = 0.5/jac* (y13*y13+x13*x13);
  ske[1][2] = 0.5/jac* (y21*y13+x21*x13);
  ske[2][0] = ske[0][2];
  ske[2][1] = ske[1][2];
  ske[2][2] = 0.5/jac* (y21*y21+x21*x21);
  
  const double xm    = (xc[i1+0] + xc[i2+0] + xc[i3+0])/3.0,
               ym    = (xc[i1+1] + xc[i2+1] + xc[i3+1])/3.0;
  fe[0] = fe[1] = fe[2] = 0.5*jac* FunctF(xm,ym)/3.0;
}

/* 
	general routine for lin. triangular elements,
	symm. matrix
	node numbering in element: lex. forward
*/
void AddElem(const int ial[3], const double ske[3][3], const double fe[3], 
	double sk[], const int id[], const int ik[], double f[])
{
  for (int i=0; i<3; ++i)
  {
    const int ii  = ial[i],
              id1 = (ii==0) ? -1 : id[ii-1],
              id2 = id[ii];
          int ip  = id1;
    for (int j=i; j<3; ++j)
    {
      const int jj = ial[j];
      int btest = -1;		// FIXXXIT: boolean
      do
      { ++ip;
        btest = (ik[ip]==jj);
      }
      while (!btest && ip<id2);

      if (!btest) 
      	{ cout << "Error in AddElem: (" << ii << "," << jj << ") [" 
	       << ial[0] << "," << ial[1] << "," << ial[2] << "]\n"; 
	  exit(-123);
	}
      
      sk[ip] += ske[i][j];
    }
    f[ii] += fe[i];
  }   
}

/* 
*/
void DebugMatrix(const int nnode, const double sk[], const int id[], const int ik[])
{
  cout << "\nMatrix\n";
  int id1, id2 = -1;
  for (int i=0; i<nnode; ++i)
  {
    cout << "Row " << i << " : ";
    id1 = id2+1;
    id2 = id[i];
    for (int j=id1; j<=id2; ++j)
      cout << "[" << ik[j] << "]  " << sk[j];
    cout << endl;;
  }

  return;
}
/*
	void GetMatrix(nx, ny, xl, xr, yb, yt, sk, id, ik)

	Computes matrix for Poisson eqution. 
	a) Determine coordinates and element connectivity matrix wrt. 
		an equidistant discretization.
	b) We use a general code to calculate element matrices and 
		local stiffness matrix K. 

	IN:	nx, ny	points in x/y direction
		dx, dy  subdomain-size in x/y-direction
	OUT:	sk	symmetric matrix
		id, ik	crs pointer of sk
*/

void GetMatrix(const int nx, const int ny, 
		const double& xl, const double& xr, const double& yb, const double& yt, 
		double sk[], int id[], int ik[], double f[])
{
  const int nnode = (nx+1)*(ny+1),
            nelem = 2*nx*ny; 
  
  double *xc = new double [2*nnode];		/* coordinates	  */
  GetCoordsInRectangle(nx, ny, xl, xr, yb, yt, xc);

  int *ia = new int [3*nelem]; 			/* connectivity matrix */
  GetConnectivityInRectangle(nx, ny, ia);

  GetMatrixpointersInRectangle(nx, ny, id, ik);
  
  const double dzero=0.0;
  vdcopy(id[nnode-1], sk, 1, &dzero, 0);
  vdcopy(nnode, f, 1, &dzero, 0);
  
  double ske[3][3], fe[3];
						/* Loop over all elements */
  for (int i=0; i<nelem; ++i)
  {
    CalcElem(&ia[3*i], xc, ske, fe);
    AddElem(&ia[3*i], ske, fe, sk, id, ik, f);
  }

  DebugMatrix(nnode, sk, id, ik); 

  delete [] ia;  delete [] xc;
  return;
}
