#include <stdio.h>
#include <math.h>
#include <string.h>

/******************************************************************************/
/*M
  Vdcopy - Copy a double vector

  DESCRIPTION: 
    Copies a double vector x (stride jx) to the double vector y (stride jy). 
.n  y_i := x_j ; i=1+k*jy; j=1+k*jx;  k=0,n-1

  SYNOPSIS:
   void Vdcopy(nn, xp, jx, yp, jy)

   double           *xp,*yp;
   int           jx,jy;
   unsigned long nn;

  PARAMETERS:
.   nn  - #components to copy (I)  
.   *xp - input vector (I)
.   jx  - stride of xp (I)
.   *yp - output vector (O)
.   jy  - stride of yp (I)
M*/
/******************************************************************************/

void vdcopy(nn, xp, jx, yp, jy)

   double        *xp,*yp;
   int           jx,jy;
   unsigned long nn;

{
   double        *x=xp,*y=yp;
   int           ix= jx,iy= jy;
   unsigned long n=nn,i,m;

   if ((ix==1)&&(iy==1)) 
   { 
      memcpy(xp,yp,(size_t)8*n); 
      return; 
   }; 
   
   m = n%7;
   if (m)
   {  
      for (i=1; i<=m; *x= *y, x+=ix, y+=iy, i++);
   };

   if (n<7) return;

   for (i=m+1; i<=n; i+=7)
   {
      *x= *y; x+=ix; y+=iy;
      *x= *y; x+=ix; y+=iy;
      *x= *y; x+=ix; y+=iy;
      *x= *y; x+=ix; y+=iy;
      *x= *y; x+=ix; y+=iy;
      *x= *y; x+=ix; y+=iy;
      *x= *y; x+=ix; y+=iy;
   };
   return;
}


/*M
  Vdplus - Adds two vectors into a third one.

  DESCRIPTION: 
    Adds the components of the double vector x (stride jx) an the double 
    vector y (stride jy). Te resulting vector z has stride jz.
.n  z_k := y_i + x_j ; i=1+l*jy; j=1+l*jx; k=1+l*jz;  l=0,n-1

  SYNOPSIS:
   void Vdplus(nn, xp, jx, yp, jy, zp, jz)

   double        *xp,*yp,*zp;
   int           jx,jy,yz;
   unsigned long nn;

  PARAMETERS:
.   nn  - #components (I)  
.   *xp - first vector (I)
.   jx  - stride of xp (I)
.   *yp - second vector (I)
.   jy  - stride of yp (I)
.   *zp - resulting vector (O)
.   jz  - stride of zp (I)
M*/
/******************************************************************************/

void vdplus(nn, xp, jx, yp, jy, zp, jz)

   double        *xp,*yp,*zp;
   int           jx,jy,jz;
   unsigned long nn;

{
   double        *x=xp,*y=yp,*z=zp;
   int           ix=jx,iy=jy,iz=jz;
   unsigned long n=nn,i,m;

   if ((ix==1)&&(iy==1)&&(iz==1)) goto m1;

   for (i=1; i<=n; *x= *y+*z, x+=ix, y+=iy, z+=iz, i++);
   return;

m1:m = n%4;
   if (m)
   {
      for (i=1; i<=m; *x= *y+*z, x++, y++, z++, i++);
   };

   if (n<4) return;

   for (i=m+1; i<=n; i+=4, x+=4, y+=4,z+=4)
   { 
      *x= *y+*z;
      *(x+1)= *(y+1)+*(z+1);
      *(x+2)= *(y+2)+*(z+2);
      *(x+3)= *(y+3)+*(z+3);
   };
   return;
}

void vddiv(nn, xp, jx, yp, jy, zp, jz)

   double        *xp,*yp,*zp;
   int           jx,jy,jz;
   unsigned long nn;

{
   double        *x=xp,*y=yp,*z=zp;
   int           ix=jx,iy=jy,iz=jz;
   unsigned long n=nn,i,m;

   if ((ix==1)&&(iy==1)&&(iz==1)) goto m1;

   for (i=1; i<=n; *x= *y/ *z, x+=ix, y+=iy, z+=iz, i++);
   return;

m1:m = n%4;
   if (m)
   {
      for (i=1; i<=m; *x= *y/ *z, x++, y++, z++, i++);
   };

   if (n<4) return;

   for (i=m+1; i<=n; i+=4, x+=4, y+=4,z+=4)
   { 
      *x= *y/ *z;
      *(x+1)= *(y+1)/ *(z+1);
      *(x+2)= *(y+2)/ *(z+2);
      *(x+3)= *(y+3)/ *(z+3);
   };
   return;
}

void vdmult(nn, xp, jx, yp, jy, zp, jz)

   double        *xp,*yp,*zp;
   int           jx,jy,jz;
   unsigned long nn;

{
   double        *x=xp,*y=yp,*z=zp;
   int           ix=jx,iy=jy,iz=jz;
   unsigned long n=nn,i,m;

   if ((ix==1)&&(iy==1)&&(iz==1)) goto m1;

   for (i=1; i<=n; *x= *y* *z, x+=ix, y+=iy, z+=iz, i++);
   return;

m1:m = n%4;
   if (m)
   {
      for (i=1; i<=m; *x= *y* *z, x++, y++, z++, i++);
   };

   if (n<4) return;

   for (i=m+1; i<=n; i+=4, x+=4, y+=4,z+=4)
   { 
      *x= *y**z;
      *(x+1)= *(y+1)* *(z+1);
      *(x+2)= *(y+2)* *(z+2);
      *(x+3)= *(y+3)* *(z+3);
   };
   return;
}
/******************************************************************************/
/*M
  Dscapr - inner product of two vectors

  DESCRIPTION: 
    Calculates the inner product of two double precision vectors (stride 1). 
.n  d := sum (x_i*y_i) ; i=1,n

  SYNOPSIS:
   double Dscapr(n, xp, yp)

   double        *xp,*yp;
   unsigned long n;

  PARAMETERS:
.   n  - #components of the vectors (I)  
.   xp - vector (I)
.   yp - vector (I)

M*/
/******************************************************************************/

double dscapr(n, xp, yp)

   double        *xp,*yp;
   unsigned long n;

{
   double        *x=xp,*y=yp,s;
   unsigned long i,m;

   s=0.0e0;
   m = n%5;
   if (m)
   { 
      for (i=1; i<=m; s += *x * *y, i++, x++, y++);
   };
   if (n<5) return(s);
   for (i=m+1; i<=n; i+=5, x+=5, y+=5)
   {
      s += *x * *y + *(x+1) * *(y+1) + *(x+2) * *(y+2) + *(x+3) * *(y+3) + *(x+4) * *(y+4);
   };
   return(s);
}


void vdaxpy(np, xp, yp, ap, zp )

   double        *xp,*yp,*zp,ap;
   unsigned long np;

{
   double        *x=xp,*y=yp,*z=zp;
   register double a;
   register unsigned long n = np,i,m;

   a = ap;
   m = n%4;
   if (m)
   {
      for (i=1; i<=m; *x= *y + a * *z, i++, x++, y++, z++);
   };
   if (n<4) return;
   for (i=m+1; i<=n; i+=4, x+=4, y+=4, z+=4)
   {
      *x = *y + a * *z;
      *(x+1) = *(y+1) + a * *(z+1);
      *(x+2) = *(y+2) + a * *(z+2);
      *(x+3) = *(y+3) + a * *(z+3);
   };
   return;
}




void vicopy(nn, xp, jx, yp, jy)

   int        *xp,*yp;
   int           jx,jy;
   unsigned long nn;

{
   int        *x=xp,*y=yp;
   int           ix= jx,iy= jy;
   unsigned long n=nn,i,m;

   if ((ix==1)&&(iy==1)) 
   { 
      memcpy(xp,yp,(size_t)4*n); 
      return; 
   }; 
   
   m = n%7;
   if (m)
   {  
      for (i=1; i<=m; *x= *y, x+=ix, y+=iy, i++);
   };

   if (n<7) return;

   for (i=m+1; i<=n; i+=7)
   {
      *x= *y; x+=ix; y+=iy;
      *x= *y; x+=ix; y+=iy;
      *x= *y; x+=ix; y+=iy;
      *x= *y; x+=ix; y+=iy;
      *x= *y; x+=ix; y+=iy;
      *x= *y; x+=ix; y+=iy;
      *x= *y; x+=ix; y+=iy;
   };
   return;
}
