#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <set>
using namespace std;

#include "mesh.h"

// -----------------------------------------------------------------------------
Mesh::Mesh()
  : eptr_(0), eind_(0), nnodes_(-1)
{}
// -----------------------------------------------------------------------------
Mesh::Mesh(const vector< vector<int> > &con)
  : eptr_(con.size()+1), eind_(0), nnodes_(-1)
{
 eptr_[0] = 0;
 for (unsigned int i=0; i<con.size(); ++i)
 {
   const vector<int> &ni = con[i];
   eptr_[i+1] = eptr_[i] + ni.size();
 }
 
 eind_.resize( eptr_.back() );	// Allocate memory for element description
 
 int k=0;
 for (unsigned int i=0; i<con.size(); ++i)
 {
   const vector<int> &ni = con[i];
   for (unsigned int j=0; j<ni.size(); ++j, ++k)
   {
     eind_[k] = ni[j];
   }
 }

 return;
}

// -----------------------------------------------------------------------------
Mesh::Mesh(const Mesh& orig)
  : eptr_(orig.eptr_), eind_(orig.eind_), nnodes_(orig.nnodes_)
{}

// -----------------------------------------------------------------------------
Mesh::Mesh(const int proc, const int elmdist[], const Mesh& orig)
  : eptr_(0), eind_(0), nnodes_(-1)
{
  int ibeg, iend;
//    Determine  number of elements and lenght of eind_ wrt. elmdist[]
  int nelem=0, nsize=0;
  
  iend = orig.eptr_[0];
  for (int i=0; i<orig.numElements(); ++i)
  {
    ibeg = iend;
    iend = orig.eptr_[i+1];
    if ( elmdist[i] == proc )
    {
       ++nelem;
       nsize += iend-ibeg;
    }
  }
  
//  Allocate memory and copy the local (proc)  data
  eptr_.resize(nelem+1);
  eind_.resize(nsize);

  int count = 0;  // local element counter
  int k = 0;      // index for eind_ array
  eptr_[count] = k;
  
  iend = orig.eptr_[0];
  for (int i=0; i<orig.numElements(); ++i)
  {
    ibeg = iend;
    iend = orig.eptr_[i+1];
    if ( elmdist[i] == proc )
    {
       for (int j=ibeg; j<iend; ++j, ++k)
           eind_[k] = orig.eind_[j];
       ++count;
       eptr_[count] = k;
    }
  }
  
  return;
}
// -----------------------------------------------------------------------------

Mesh::Mesh(const string &filename)
  : eptr_(0), eind_(0), nnodes_(-1)
{
 ifstream infile(filename.c_str());
 
 int nelem;
 infile >> nelem;
 
 //GH:   avoid the reallocation within push_back() !!
 eptr_.resize(nelem+1);
 eptr_[0] = 0;
 for (int i=1; i<=nelem; ++i)
 {
   int nl;
   infile >> nl;
   eptr_[i] = eptr_[i-1] + nl;

   for (int k=0; k<nl; ++k)
   {
       int ik;
       infile >> ik;
       eind_.push_back(ik);
   }
 }

 return;
}
// -----------------------------------------------------------------------------
Mesh::~Mesh()
{}

// -----------------------------------------------------------------------------
Mesh& Mesh::operator=(const Mesh& rhs)
{
 if (this != &rhs)
 {
   eptr_ = rhs.eptr_;
   eind_ = rhs.eind_;
   nnodes_ = rhs.nnodes_;
 }
 return *this;
}

// -----------------------------------------------------------------------------

int Mesh::numNodes() const
{
  if ( nnodes_ < 0 )
  {
//  Ok, now we have to determine the number of nodes
//            The next line works only for an complete sequence of nodes
/*    nnodes_ = *max_element(eind_.begin(),eind_.end()) + 1;*/
//            This version works always
    set<int> ss(eind_.begin(),eind_.end());
    nnodes_ = ss.size();
  }
  return nnodes_;

}
// -----------------------------------------------------------------------------

void Mesh::ConvertToLocalNodeNumbering(vector<int> &loc2glob)
{
 set<int> ss(eind_.begin(),eind_.end());
 loc2glob = vector<int>(ss.begin(),ss.end());
 
 // Convert the element connectivity from global to local node numbers
 const int nglob = *max_element(loc2glob.begin(),loc2glob.end());
 vector<int> g2l(nglob, -1);    // global to local mapping
 for (unsigned int i=0; i<loc2glob.size(); ++i)
      g2l[ loc2glob[i] ] = i;
 
 // Now, do the renumbering of element connectivity
 for (unsigned int i=0; i<eind_.size(); ++i)
      eind_[i] = g2l[ eind_[i] ];
 
 return;
}
// -----------------------------------------------------------------------------

Mesh  Mesh::DualMesh() const
{
// Estimate the Memory needed and reserve the memory
 const int MAX_NUMBER = eind_.size()/numNodes() *2;
 vector<int> tmp(0);
 tmp.reserve(MAX_NUMBER);
 vector< vector<int> > n2e(numNodes(),tmp);

 // Assign elements to nodes
 int ibeg, iend;
 iend = eptr_[0];
 for (int i=0; i<numElements(); ++i)
  {
    ibeg = iend;
    iend = eptr_[i+1];
    for (int k=ibeg; k<iend; ++k)
    {
       n2e[eind_[k]].push_back(i);
    }
  }
 
 // sort the list of elements per node
 for (int i=0; i<numNodes(); ++i)
 {
   vector<int> &ni = n2e[i];
   sort( ni.begin(), ni.end() );
/*   cout << "node  " << i << " : ";
   for (unsigned int k=0; k<ni.size(); ++k)
       cout << "  " << ni[k];
   cout << endl;*/
 }
 
 //dual = Mesh(n2e);

 return Mesh(n2e);
}

// -----------------------------------------------------------------------------

vector< vector<int> > Mesh::GetMeshArrays() const
{
 vector< vector<int> > e2n;

 e2n.reserve(numElements());
  // Assign array elements to vector elements
 vector<int>::const_iterator ibeg, iend;
 iend = eind_.begin()+eptr_[0];
 for (int i=0; i<numElements(); ++i)
  {
    ibeg = iend;
    iend = ibeg+eptr_[i+1]-eptr_[i];
    e2n.push_back( vector<int>(ibeg, iend) );
  }
  
 return e2n;
 
}



// -----------------------------------------------------------------------------
//##############################################################################

ostream & operator<<(ostream  & s, const Mesh & orig)
{
 s << "\nMesh with  " <<  orig.numElements() << "  elements and   "
   << orig.numNodes() << "  nodes\n\n";
 
 s << "Mesh pointer:\n";
 for (int i=0; i<=orig.numElements(); ++i)
    s << "  " << orig.eptr_[i];
 s << endl << endl;
 
 s << "Elements:\n";
 for (int i=0; i<orig.numElements(); ++i)
  {
    s << i << " :";
    for (int k=orig.eptr_[i]; k<orig.eptr_[i+1]; ++k)
    {
       s << "   " << orig.eind_[k];
    }
    s << endl;
  }
 return s;
}
// -----------------------------------------------------------------------------



ostream & operator<<(ostream  & s, const vector< vector<int> > & orig)
{
 s << "\nArray with  " <<  orig.size() << "  subarrays\n\n";
 
 for (unsigned int i=0; i<orig.size(); ++i)
 {
   const vector<int> &ni = orig[i];
   s << "node " << i << " : ";
   for (unsigned int j=0; j<ni.size(); ++j)
         s << "  " << ni[j];
   s << endl;
 }
 return s;
}
// -----------------------------------------------------------------------------

bool smaller(const vector<int> & lhs, const vector<int> & rhs)
{
  bool bs;
  if ( lhs.size() != rhs.size() )
  {
    bs = (lhs.size() < rhs.size());
  }
  else
  {
    bs = (lhs < rhs);
  }
  
  return bs;
}

